# -*- coding: utf-8 -*-
#
#   This file is part of the magpurify package, available at:
#   https://github.com/snayfach/MAGpurify
#
#   Magpurify is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.

import os
import subprocess as sp
import sys
from Bio import Seq, SeqIO


def add_tmp_dir(args):
    tmp_dir = f"{args['out']}/{args['program']}"
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
    args["tmp_dir"] = tmp_dir


def check_input(args):
    if not os.path.exists(args["fna"]):
        error = f"\nInput file not found: {args['fna']}\n"
        sys.exit(error)


def check_dependencies(programs):
    for program in programs:
        if not exists_on_env_path(program):
            error = f"\nRequired program '{program}' not found\n"
            error += "Make sure this program has been installed and added to your PATH\n"
            sys.exit(error)


def exists_on_env_path(program):
    " Check whether program exists in PATH and is executable"
    for dir in os.environ["PATH"].split(os.pathsep):
        fpath = dir + "/" + program
        if os.path.exists(fpath) and os.access(fpath, os.X_OK):
            return True
    return False


def check_database(args):
    if args["db"] is None and "MAGPURIFYDB" in os.environ:
        args["db"] = os.environ["MAGPURIFYDB"]
    if args["db"] is None:
        error = "\nError: No reference database specified\n"
        error += "Use the --db argument to specify a database,\n"
        error += "Or set the MAGPURIFYDB environmental variable: export MAGPURIFYDB=/path/to/MAGpurify_db_v1.0.0\n"
        sys.exit(error)
    if not os.path.isdir(args["db"]):
        error = f"\nError: Specified reference database does not exist: {args['db']}\n"
        error += (
            "\nCheck that you've entered the path correctly and the database exists\n"
        )
        sys.exit(error)


def reverse_complement(sequence):
    rc = Seq.Seq(sequence).reverse_complement()
    return str(rc)


def run_process(command):
    process = sp.Popen(command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    out, err = process.communicate()
    if process.returncode != 0:
        err_message = (
            f"\nError encountered executing:\n{command}\n\nError message:\n{err}\n"
        )
        sys.exit(err_message)
    return out, err


def parse_last(inpath):
    fields = [
        "qid",
        "tid",
        "pid",
        "aln",
        "mis",
        "gap",
        "qstart",
        "qend",
        "tstart",
        "tend",
        "eval",
        "score",
        "qlen",
        "tlen",
    ]
    for line in open(inpath):
        if line[0] == "#":
            continue
        else:
            values = line.rstrip().split()
            d = dict([(f, v) for f, v in zip(fields, values)])
            d["qcov"] = float(d["aln"]) / float(d["qlen"])
            d["tcov"] = float(d["aln"]) / float(d["tlen"])
            yield d


def parse_blast(input, type="file"):
    formats = [
        ("qname", str),
        ("tname", str),
        ("pid", float),
        ("aln", int),
        ("mis", int),
        ("gap", int),
        ("qstart", int),
        ("qend", int),
        ("tstart", int),
        ("tend", int),
        ("evalue", float),
        ("bitscore", float),
        ("qlen", int),
        ("tlen", int),
    ]
    if type == "file":
        lines = open(input).read().rstrip("\n").split("\n")
    else:
        lines = input.rstrip("\n").split("\n")
    if lines == [""]:
        return
    for line in lines:
        values = line.split("\t")
        record = dict([(f[0], f[1](v)) for f, v in zip(formats, values)])
        record["qcov"] = 100 * record["aln"] / float(record["qlen"])
        record["tcov"] = 100 * record["aln"] / float(record["tlen"])
        yield record


def parse_mash(fpath):
    out = open(fpath).read()
    if len(out) > 0:
        lines = out.rstrip("\n").split("\n")
        fields = ["query", "target", "dist", "pvalue", "fraction"]
        formats = [str, str, float, float, str]
        for line in lines:
            values = line.split()
            rec = dict([(f, m(v)) for f, m, v in zip(fields, formats, values)])
            yield rec


def parse_hmmsearch(fpath):
    with open(fpath) as infile:
        fields = [
            "tname",
            "tacc",
            "tlen",
            "qname",
            "qacc",
            "qlen",
            "evalue",
            "score",
            "bias",
            "ndom",
            "tdom",
            "c-evalue",
            "i-evalue",
            "domscore",
            "dombias",
            "hmmfrom",
            "hmmto",
            "alifrom",
            "alito",
            "envfrom",
            "envto",
            "prob",
            "tdesc",
        ]
        formts = [
            str,
            str,
            int,
            str,
            str,
            int,
            float,
            float,
            float,
            int,
            int,
            float,
            float,
            float,
            float,
            int,
            int,
            int,
            int,
            int,
            int,
            float,
            str,
        ]
        for line in infile:
            if line[0] == "#":
                continue
            values = line.rstrip("\n").split(None, 22)
            r = dict(
                [
                    (field, format(value))
                    for field, format, value in zip(fields, formts, values)
                ]
            )
            r["tcov"] = float(r["alito"] - r["alifrom"] + 1) / r["tlen"]  # target is gene
            r["qcov"] = float(r["hmmto"] - r["hmmfrom"] + 1) / r["qlen"]  # query is hmm
            yield r


def fetch_hmm_best_hits(fpath):
    gene_to_aln = {}
    for aln in parse_hmmsearch(fpath):
        if aln["tname"] not in gene_to_aln:
            gene_to_aln[aln["tname"]] = aln
        elif aln["score"] > gene_to_aln[aln["tname"]]["score"]:
            gene_to_aln[aln["tname"]] = aln
    return gene_to_aln


def run_coverm(bam_list, out_dir, threads):
    cmd = "coverm contig "
    cmd += f"--threads {threads} "
    cmd += "--contig-end-exclusion 75 "
    cmd += "--min-read-percent-identity 0.97 "
    cmd += f"--bam-files {' '.join(bam_list)} "
    cmd += f"> {out_dir}/coverage.tsv"
    out, err = run_process(cmd)


def run_prodigal(fna_path, out_dir):
    cmd = "prodigal "
    cmd += f"-i {fna_path} "  # input fna
    cmd += f"-a {out_dir}/genes.faa "  # protein seqs
    cmd += f"-d {out_dir}/genes.ffn "  # nucleotide seqs
    cmd += f"-o {out_dir}/genes.out "  # prodigal output
    cmd += "> /dev/null"  # prodigal output
    out, err = run_process(cmd)


def run_lastal(db_dir, out_dir, threads=1, seed_freq=10):
    cmd = "lastal -p BLOSUM62 -P 1 -f blasttab+ "
    cmd += f"-m {seed_freq} "
    cmd += f"{db_dir}/clade-markers/markers.faa "
    cmd += f"{out_dir}/genes.faa "
    cmd += f"-P {threads} "
    cmd += f"> {out_dir}/genes.m8 "
    out, err = run_process(cmd)


def run_hmmsearch(db_dir, in_path, out_dir, threads=1):
    cmd = "hmmsearch "
    cmd += "--noali "
    cmd += f"--domtblout {out_dir}/phyeco.hmmsearch "
    cmd += f"--cpu {threads} "
    cmd += "--cut_ga "
    cmd += f"{db_dir}/phylo-markers/PhyEco.hmm "
    cmd += f"{in_path}/genes.faa "
    out, err = run_process(cmd)


def run_blastp(db_path, query_path, out_path, threads=1, max_targets=1, qcov=40):
    cmd = "blastp "
    cmd += f"-db {db_path} "
    cmd += f"-query {query_path} "
    cmd += f"-out {out_path} "
    cmd += "-outfmt '6 std qlen slen' "
    cmd += f"-num_threads {threads} "
    cmd += f"-max_target_seqs {max_targets} "
    cmd += f"-qcov_hsp_perc {qcov} "
    cmd += "-max_hsps 1 "
    out, err = run_process(cmd)


def run_blastn(db_path, query_path, out_path, threads=1, max_targets=1, qcov=40):
    cmd = "blastn -task blastn "
    cmd += f"-db {db_path} "
    cmd += f"-query {query_path} "
    cmd += f"-out {out_path} "
    cmd += "-outfmt '6 std qlen slen' "
    cmd += f"-num_threads {threads} "
    cmd += f"-max_target_seqs {max_targets} "
    cmd += f"-qcov_hsp_perc {qcov} "
    cmd += "-max_hsps 1 "
    out, err = run_process(cmd)


def parse_fasta(path):
    with open(path) as file:
        for record in SeqIO.parse(file, "fasta"):
            id = record.id
            seq = str(record.seq).upper()
            yield id, seq



# -*- coding: utf-8 -*-
#
#   This file is part of the magpurify package, available at:
#   https://github.com/snayfach/MAGpurify
#
#   Magpurify is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import csv
import os
import sys
from collections import Counter
from magpurify import utilities


def fetch_args(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="phylo-markers")
    parser.add_argument(
        "fna",
        type=str,
        help="Path to input genome in FASTA format"
    )
    parser.add_argument(
        "out",
        type=str,
        help="Output directory to store results and intermediate files",
    )
    parser.add_argument(
        "--db",
        type=str,
        help="Path to reference database. By default, the MAGPURIFYDB environmental variable is used",
    )
    parser.add_argument(
        "--continue",
        action="store_true",
        default=False,
        help="Go straight to quality estimation and skip all previous steps",
    )
    parser.add_argument(
        "--max_target_seqs",
        type=int,
        default=1,
        help="Maximum number of targets reported in BLAST table",
    )
    parser.add_argument(
        "--cutoff_type",
        choices=["strict", "sensitive", "none"],
        default="strict",
        help="Use strict or sensitive %%ID cutoff for taxonomically annotating genes",
    )
    parser.add_argument(
        "--seq_type",
        choices=["dna", "protein", "both", "either"],
        default="protein",
        help="Choose to search genes versus DNA or protein database",
    )
    parser.add_argument(
        "--hit_type",
        choices=["all_hits", "top_hit"],
        default="top_hit",
        help="Transfer taxonomy of all hits or top hit per gene",
    )
    parser.add_argument(
        "--exclude_clades",
        type=str,
        nargs="+",
        help="List of clades to exclude (ex: s__1300164.4)",
    )
    parser.add_argument(
        "--bin_fract",
        type=float,
        default=0.7,
        help="Min fraction of genes in bin that agree with consensus taxonomy for bin annotation",
    )
    parser.add_argument(
        "--contig_fract",
        type=float,
        default=1.0,
        help="Min fraction of genes in that disagree with bin taxonomy for filtering",
    )
    parser.add_argument(
        "--allow_noclass",
        action="store_true",
        default=False,
        help="Allow a bin to be unclassfied and flag any classified contigs",
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=1,
        help="Number of CPUs to use",
    )


def extract_homologs(tmp_dir):
    # create outdir
    seqs_dir = f"{tmp_dir}/markers"
    if not os.path.isdir(seqs_dir):
        os.makedirs(seqs_dir)
    # fetch best hits from hmmsearch
    gene_to_aln = utilities.fetch_hmm_best_hits(f"{tmp_dir}/phyeco.hmmsearch")
    # open output files
    outfiles = {}
    marker_ids = set(aln["qacc"] for aln in list(gene_to_aln.values()))
    for marker_id in marker_ids:
        outfiles[marker_id] = {}
        outfiles[marker_id]["ffn"] = open(f"{seqs_dir}/{marker_id}.ffn", "w")
        outfiles[marker_id]["faa"] = open(f"{seqs_dir}/{marker_id}.faa", "w")
    # write seqs
    for ext in ["ffn", "faa"]:
        in_path = f"{tmp_dir}/genes.{ext}"
        for id, seq in utilities.parse_fasta(in_path):
            if id in gene_to_aln:
                marker_id = gene_to_aln[id]["qacc"]
                seq = seq.rstrip("*")
                outfiles[marker_id][ext].write(">" + id + "\n" + seq + "\n")
    # close files
    for marker_id in outfiles:
        outfiles[marker_id]["ffn"].close()
        outfiles[marker_id]["faa"].close()


def align_homologs(db_dir, tmp_dir, seq_type, threads):
    aln_dir = f"{tmp_dir}/alns"
    if not os.path.exists(aln_dir):
        os.makedirs(aln_dir)
    seq_files = os.listdir(f"{tmp_dir}/markers")
    for file_index, seq_file in enumerate(seq_files):
        marker_id, ext = seq_file.split(".")
        if ext == "ffn" and seq_type == "protein":
            continue
        elif ext == "faa" and seq_type == "dna":
            continue
        program = 'blastp' if ext == 'faa' else 'blastn'
        db_path = f"{db_dir}/phylo-markers/{program}/{marker_id}"
        query_path = f"{tmp_dir}/markers/{marker_id}.{ext}"
        out_path = f"{tmp_dir}/alns/{marker_id}.{ext}.m8"
        utilities.run_blastp(db_path, query_path, out_path, threads)


class Bin:
    def __init__(self):
        self.cons_taxon = None
        self.cons_fract = None
        self.cons_count = None
        self.genes = None

    def exclude_clades(self, clades):
        # loop over genes in bin
        for gene in list(self.genes.values()):
            # keep track of which annotations to exclude for each gene
            exclude_indexes = []
            # loop over annotations for each gene
            for index, annotation in enumerate(gene.annotations):
                is_match = any(c in annotation.taxon for c in clades)
                if is_match:
                    exclude_indexes.append(index)
            # delete annotations
            for index in exclude_indexes[::-1]:
                del gene.annotations[index]

    def only_keep_top_hits(self):
        for gene in list(self.genes.values()):
            if len(gene.annotations) > 1:
                max_score = max(a.score for a in gene.annotations)
                gene.annotations = [a for a in gene.annotations if a.score == max_score]

    def classify_taxonomy(self, allow_noclass=False, min_fraction=0.5):
        ranks = ["s", "g", "f", "o", "c", "p"]
        for rank_index, rank in enumerate(ranks):
            # get list of taxa across all genes
            # never count a taxon 2x per gene
            gene_taxa = []
            for gene in list(self.genes.values()):
                # do not count unclassified
                if not allow_noclass:
                    taxa = list(set(a.taxon[rank_index] for a in gene.annotations
                                                    if a.taxon[rank_index]))
                else:
                    taxa = list(set(a.taxon[rank_index] for a in gene.annotations))
                gene_taxa += taxa
            # skip rank where there are no annotations
            if not gene_taxa:
                continue
            # get most common annotation
            cons_taxon, cons_count = Counter(gene_taxa).most_common()[0]
            # determine if enough genes have consensus annotation
            cons_fract = cons_count / float(len(self.genes))
            if cons_fract >= min_fraction:
                self.cons_taxon = cons_taxon
                self.cons_fract = cons_fract
                self.cons_count = cons_count
                self.rank = rank
                self.rank_index = rank_index
                return


class Marker:
    def __init__(self):
        self.id = None


class Gene:
    def __init__(self):
        self.id = None
        self.marker = None
        self.contig = None
        self.annotations = []


class Annotation:
    def __init__(self):
        self.taxon = [None] * 6
        self.score = None

    def add_taxon(self, genome_taxon, rank_index):
        for i in range(rank_index, 6):
            self.taxon[i] = genome_taxon.split(";")[i]


class Contig:
    def __init__(self):
        self.id = None
        self.genes = []
        self.genes_agree = 0
        self.genes_conflict = 0
        self.flagged = None

    def compare_taxonomy(self, bin):
        for gene in self.genes:
            agrees = [
                bin.cons_taxon == annotation.taxon[bin.rank_index]
                for annotation in gene.annotations
            ]

            if any(agrees):
                self.genes_agree += 1
            else:
                self.genes_conflict += 1

    def flag(self, min_fract):
        if len(self.genes) == 0:
            self.flagged = None
        elif self.genes_conflict / float(len(self.genes)) >= min_fract:
            self.flagged = True
        else:
            self.flagged = False


def flag_contigs(db_dir, tmp_dir, args):
    # step 0. read in reference data files
    # cutoffs
    cutoffs = {}
    cutoffs_path = f"{db_dir}/phylo-markers/max_fscores.tsv"
    reader = csv.DictReader(open(cutoffs_path), delimiter="\t")
    for r in reader:
        key = (r["marker_id"], r["seq_type"], r["score_type"], r["taxlevel"])
        value = {"sensitive": r["cutoff_lower"], "strict": r["cutoff_upper"], "none": 0.0}
        cutoffs[key] = value
    taxonomy_path = f"{db_dir}/phylo-markers/genome_taxonomy.tsv"
    reader = csv.DictReader(open(taxonomy_path), delimiter="\t")
    # taxonomy
    taxonomy = {r["genome_id"]: r["taxonomy"] for r in reader}
    # clustered seqs
    clusters = {}
    for type in ["ffn", "faa"]:
        clusters[type] = {}
        for file in os.listdir(f"{db_dir}/phylo-markers/{type}"):
            if file.split(".")[-1] == "uc":
                with open(f"{db_dir}/phylo-markers/{type}/{file}") as f:
                    for l in f:
                        v = l.rstrip().split()
                        rep_id = v[-1]
                        seq_id = v[-2]
                        if v[0] == "S":
                            clusters[type][seq_id] = [seq_id]
                        elif v[0] == "H":
                            clusters[type][rep_id].append(seq_id)
    # step 1. determine if bin is archaea or bacteria; initialize domain-level markers
    # to do: normalize counts by site of marker gene sets
    marker_ids = set([])
    counts = {"bacteria": 0, "archaea": 0}
    for aln_file in os.listdir(f"{tmp_dir}/alns"):
        marker_id, seq_type, ext = aln_file.split(".")
        marker_ids.add(marker_id)
    for marker_id in marker_ids:
        if "B" in marker_id:
            counts["bacteria"] += 1
        elif "A" in marker_id:
            counts["archaea"] += 1
    domain = "bacteria" if counts["bacteria"] >= counts["archaea"] else "archaea"
    markers = {}
    for marker_id in marker_ids:
        if domain == "bacteria" and "B" in marker_id:
            markers[marker_id] = Marker()
            markers[marker_id].id = marker_id
            markers[marker_id].genes = []
        elif domain == "archaea" and "A" in marker_id:
            markers[marker_id] = Marker()
            markers[marker_id].id = marker_id
            markers[marker_id].genes = []
    # step 2. initialize marker genes found in bin
    bin = Bin()
    bin.genes = {}
    hmm_path = f"{tmp_dir}/phyeco.hmmsearch"
    for gene_id, aln in list(utilities.fetch_hmm_best_hits(hmm_path).items()):
        if aln["qacc"] not in markers:
            continue
        gene = Gene()
        gene.id = gene_id
        gene.contig = gene_id.rsplit("_", 1)[0]
        gene.marker = aln["qacc"]
        bin.genes[gene_id] = gene
        markers[aln["qacc"]].genes.append(gene)
    # annotate genes
    #    fetch all non-redundant taxonomic annotations for each gene
    seq_types = None
    if args["seq_type"] in ["both", "either"]:
        seq_types = ["ffn", "faa"]
    elif args["seq_type"] == "protein":
        seq_types = ["faa"]
    else:
        seq_types = ["ffn"]
    for seq_type in seq_types:
        for marker_id in markers:
            aln_path = tmp_dir + "/alns/" + marker_id + "." + seq_type + ".m8"
            for aln in utilities.parse_blast(aln_path):
                # fetch all unique taxonomies for target sequence
                # a sequence can have multiple taxonomies if it was clustered with another sequence
                genome_taxa = []
                for target_id in clusters[seq_type][aln["tname"]]:
                    genome_id = target_id.split("_")[0]
                    if (
                        genome_id not in taxonomy
                        or taxonomy[genome_id] in genome_taxa
                    ):
                        continue
                    else:
                        genome_taxa.append(taxonomy[genome_id])
                # loop over ranks; stop when gene has been annotated
                for rank_index, rank in enumerate(["s", "g", "f", "o", "c", "p"]):
                    if seq_type == "ffn" and rank != "s":
                        if args["seq_type"] == "either":
                            continue
                    elif seq_type == "faa" and rank == "s":
                        if args["seq_type"] == "either":
                            continue
                    # get minimum % identity cutoff for transfering taxonomy
                    #   if cutoff_type is None, indicates that no cutoff should be used
                    min_pid = cutoffs[marker_id, seq_type, "pid", rank][
                        args["cutoff_type"]
                    ]
                    if float(aln["pid"]) < float(min_pid):
                        continue
                    # add taxonomy
                    for genome_taxon in genome_taxa:
                        annotation = Annotation()
                        annotation.add_taxon(genome_taxon, rank_index)
                        annotation.score = float(aln["bitscore"])
                        bin.genes[aln["qname"]].annotations.append(annotation)
                    # stop when gene has been annotated at lowest rank
                    break
    # optionally remove annotations matching <exclude_clades>
    if args["exclude_clades"] is not None:
        bin.exclude_clades(args["exclude_clades"])
    # optionally take top hit only
    if args["hit_type"] == "top_hit":
        bin.only_keep_top_hits()
    # create None annotations for unannotated genes
    for gene in bin.genes.values():
        if len(gene.annotations) == 0:
            gene.annotations.append(Annotation())
    # classify bin
    bin.classify_taxonomy(args["allow_noclass"], args["bin_fract"])
    # flag contigs with discrepant taxonomy
    bin.contigs = {}
    for id, seq in utilities.parse_fasta(args["fna"]):
        bin.contigs[id] = Contig()
        bin.contigs[id].id = id
        bin.contigs[id].length = len(seq)
    for gene in bin.genes.values():
        bin.contigs[gene.contig].genes.append(gene)
    if bin.cons_taxon is not None:
        for contig in bin.contigs.values():
            contig.compare_taxonomy(bin)
            contig.flag(args["contig_fract"])
    return [contig.id for contig in bin.contigs.values() if contig.flagged]


def main(args):
    utilities.add_tmp_dir(args)
    utilities.check_input(args)
    utilities.check_dependencies(["prodigal", "hmmsearch", "blastp", "blastn"])
    utilities.check_database(args)
    print("\u001b[1m" + "• Calling genes with Prodigal" + "\u001b[0m")
    utilities.run_prodigal(args["fna"], args["tmp_dir"])
    print(f"  all genes: {args['tmp_dir']}/genes.[ffn|faa]")
    print("\u001b[1m" + "\n• Identifying PhyEco phylogenetic marker genes with HMMER" + "\u001b[0m")
    utilities.run_hmmsearch(args["db"], args["tmp_dir"], args["tmp_dir"], args["threads"])
    extract_homologs(args["tmp_dir"])
    print(f"  hmm results: {args['tmp_dir']}/phyeco.hmmsearch")
    print(f"  marker genes: {args['tmp_dir']}/markers")
    print("\u001b[1m" + "\n• Performing pairwise BLAST alignment of marker genes against database" + "\u001b[0m")
    align_homologs(args["db"], args["tmp_dir"], args["seq_type"], args["threads"])
    print(f"  blast results: {args['tmp_dir']}/alns")
    print("\u001b[1m" + "\n• Finding taxonomic outliers" + "\u001b[0m")
    flagged = flag_contigs(args["db"], args["tmp_dir"], args)
    out = f"{args['tmp_dir']}/flagged_contigs"
    print(f"  {len(flagged)} flagged contigs: {out}")
    with open(out, "w") as f:
        for contig in flagged:
            f.write(contig + "\n")



# -*- coding: utf-8 -*-
#
#   This file is part of the magpurify package, available at:
#   https://github.com/snayfach/MAGpurify
#
#   Magpurify is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.


# -*- coding: utf-8 -*-
#
#   This file is part of the magpurify package, available at:
#   https://github.com/snayfach/MAGpurify
#
#   Magpurify is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import sys
from magpurify import phylo, clade, conspecific, tetra, gc, coverage, contam, clean


def cli():
    parser = argparse.ArgumentParser(
        description="Identify and remove incorrectly binned contigs from metagenome-assembled genomes.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("--version", action="version", version="%(prog)s 2.1.2")
    subparsers = parser.add_subparsers()
    phylo_parser = subparsers.add_parser(
        "phylo-markers",
        help="find taxonomic discordant contigs using a database of phylogenetic marker genes.",
        description="Find taxonomic discordant contigs using a database of phylogenetic marker genes.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    phylo.fetch_args(phylo_parser)
    clade_parser = subparsers.add_parser(
        "clade-markers",
        help="find taxonomic discordant contigs using a database of clade-specific marker genes.",
        description="Find taxonomic discordant contigs using a database of clade-specific marker genes.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    clade.fetch_args(clade_parser)
    conspecific_parser = subparsers.add_parser(
        "conspecific",
        help="find contigs that fail to align to closely related genomes.",
        description="Find contigs that fail to align to closely related genomes.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    conspecific.fetch_args(conspecific_parser)
    tetra_parser = subparsers.add_parser(
        "tetra-freq",
        help="find contigs with outlier tetranucleotide frequency.",
        description="Find contigs with outlier tetranucleotide frequency.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    tetra.fetch_args(tetra_parser)
    gc_parser = subparsers.add_parser(
        "gc-content",
        help="find contigs with outlier GC content.",
        description="Find contigs with outlier GC content.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    gc.fetch_args(gc_parser)
    coverage_parser = subparsers.add_parser(
        "coverage",
        help="find contigs with outlier coverage profile.",
        description="Find contigs with outlier coverage profile.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    coverage.fetch_args(coverage_parser)
    contam_parser = subparsers.add_parser(
        "known-contam",
        help="find contigs that match a database of known contaminants.",
        description="Find contigs that match a database of known contaminants.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    contam.fetch_args(contam_parser)
    clean_parser = subparsers.add_parser(
        "clean-bin",
        help="remove putative contaminant contigs from bin.",
        description="Remove putative contaminant contigs from bin.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    clean.fetch_args(clean_parser)
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(0)
    elif len(sys.argv) == 2:
        if sys.argv[1] == "phylo-markers":
            phylo_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "clade-markers":
            clade_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "conspecific":
            conspecific_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "tetra-freq":
            tetra_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "gc-content":
            gc_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "coverage":
            coverage_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "known-contam":
            contam_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "clean-bin":
            clean_parser.print_help()
            sys.exit(0)
    args = vars(parser.parse_args())
    args["func"](args)



# -*- coding: utf-8 -*-
#
#   This file is part of the magpurify package, available at:
#   https://github.com/snayfach/MAGpurify
#
#   Magpurify is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import os
import sys
from operator import itemgetter
from magpurify import utilities


def fetch_args(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="conspecific")
    parser.add_argument("fna", type=str, help="Path to input genome in FASTA format")
    parser.add_argument(
        "out", type=str, help="Output directory to store results and intermediate files",
    )
    parser.add_argument(
        "mash_sketch", type=str, help="Path to Mash sketch of reference genomes",
    )
    parser.add_argument(
        "--mash-dist",
        type=float,
        default=0.05,
        help="Mash distance to reference genomes",
    )
    parser.add_argument(
        "--max-genomes",
        type=int,
        default=25,
        help="Max number of genomes to use",
    )
    parser.add_argument(
        "--min-genomes",
        type=int,
        default=1,
        help="Min number of genomes to use",
    )
    parser.add_argument(
        "--contig-aln",
        type=float,
        default=0.50,
        help="Minimum fraction of contig aligned to reference",
    )
    parser.add_argument(
        "--contig-pid",
        type=float,
        default=95.0,
        help="Minimum percent identity of contig aligned to reference",
    )
    parser.add_argument(
        "--hit-rate",
        type=float,
        default=0.00,
        help="Hit rate for flagging contigs",
    )
    parser.add_argument(
        "--exclude",
        nargs="+",
        default="",
        help="List of references to exclude",
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=1,
        help="Number of CPUs to use",
    )


def run_mash(mash_sketch, fna_path, tmp_dir, threads=1):
    out_path = f"{tmp_dir}/mash.dist"
    command = f"mash dist -p {threads} -d 0.25 {fna_path} {mash_sketch} > {out_path}"
    out, err = utilities.run_process(command)
    with open(tmp_dir + "/id_map.tsv", "w") as f:
        for id, rec in enumerate(utilities.parse_mash(out_path)):
            f.write(str(id) + "\t" + rec["target"] + "\n")


def find_conspecific(tmp_dir, max_dist, exclude):
    targets = []
    for rec in utilities.parse_mash(f"{tmp_dir}/mash.dist"):
        if rec["query"] == rec["target"]:
            continue
        elif rec["dist"] > max_dist:
            continue
        elif rec["pvalue"] > 1e-3:
            continue
        elif rec["target"] in exclude:
            continue
        else:
            targets.append([rec["target"], rec["dist"]])
    targets = sorted(targets, key=itemgetter(1))
    return targets


def blastn(query, target, outdir, id):
    out_path = outdir + "/" + id + ".m8"
    if not os.path.exists(out_path):
        cmd = "blastn -outfmt '6 std qlen slen' "
        cmd += "-max_target_seqs 1 -max_hsps 1 "
        cmd += f"-query {query} -subject {target} "
        out, err = utilities.run_process(cmd)
        out = out.decode("utf-8")
        open(out_path, "w").write(out)
    else:
        out = open(out_path).read()
    return out


def id_blast_hits(blast_out, min_aln, min_pid):
    blast_hits = set([])
    for rec in utilities.parse_blast(blast_out, type="string"):
        if rec["qcov"] < min_aln or rec["pid"] < min_pid:
            continue
        else:
            blast_hits.add(rec["qname"])
    return blast_hits


def align_contigs(args, genomes):
    target_to_id = {}
    for line in open(args["tmp_dir"] + "/id_map.tsv"):
        id, target = line.split()
        target_to_id[target] = id
    alignments = []
    for genome_path, mash_dist in genomes:
        blast_out = blastn(
            args["fna"], genome_path, args["tmp_dir"], target_to_id[target]
        )
        alignments.append(blast_out)
    return alignments


def find_contig_targets(args, genomes, alignments):
    contigs = dict(
        [
            (id, {"hits": 0, "len": len(seq), "genomes": []})
            for id, seq in utilities.parse_fasta(args["fna"])
        ]
    )
    for genome, alns in zip(genomes, alignments):
        hits = id_blast_hits(alns, args["contig_aln"], args["contig_pid"])
        for contig in hits:
            contigs[contig]["hits"] += 1
            contigs[contig]["genomes"].append(genome[0])
    for id in contigs:
        hit_rate = contigs[id]["hits"] / float(len(alignments))
        contigs[id]["hit_rate"] = hit_rate
    return contigs


def flag_contigs(args, contigs):
    flagged = []
    for id in contigs:
        if contigs[id]["hit_rate"] > args["hit_rate"]:
            continue
        else:
            flagged.append(id)
    return flagged


def main(args):
    utilities.add_tmp_dir(args)
    utilities.check_input(args)
    utilities.check_dependencies(["mash"])
    if not os.path.exists(args["mash_sketch"]):
        sys.exit(f"\nError: mash sketch '{args['mash_sketch']}' not found\n")
    print("\u001b[1m" + "• Finding conspecific genomes in database" + "\u001b[0m")
    run_mash(args["mash_sketch"], args["fna"], args["tmp_dir"], args["threads"])
    genomes = find_conspecific(args["tmp_dir"], args["mash_dist"], args["exclude"])
    print(f"  {len(genomes)} genomes within {args['mash_dist']} mash-dist")
    out = f"{args['tmp_dir']}/conspecific.list"
    with open(out, "w") as f:
        f.write("genome_id\tmash_dist\n")
        for genome_id, mash_dist in genomes:
            f.write(genome_id + "\t" + str(mash_dist) + "\n")
    print(f"  list of genomes: {out}")
    print(f"  mash output: {args['tmp_dir']}/mash.dist")
    if len(genomes) < args["min_genomes"]:
        sys.exit("\nError: insufficient number of conspecific genomes\n")
    if len(genomes) > args["max_genomes"]:
        print(
            "\u001b[1m"
            + f"\n• Selecting top {args['max_genomes']} most-similar genomes"
            + "\u001b[0m"
        )
        genomes = genomes[0 : args["max_genomes"]]
        out = f"{args['tmp_dir']}/conspecific_subset.list"
        with open(out, "w") as f:
            f.write("genome_id\tmash_dist\n")
            for genome_id, mash_dist in genomes:
                f.write(genome_id + "\t" + str(mash_dist) + "\n")
        print(f"  list of genomes: {out}")
    print(
        "\u001b[1m"
        + "\n• Performing pairwise alignment of contigs in bin to database genomes"
        + "\u001b[0m"
    )
    alignments = align_contigs(args, genomes)
    num_alns = sum(len(_.split("\n")) for _ in alignments)
    print(f"  total alignments: {num_alns}")
    print("\u001b[1m" + "\n• Summarizing alignments" + "\u001b[0m")
    contigs = find_contig_targets(args, genomes, alignments)
    out = f"{args['tmp_dir']}/contig_hits.tsv"
    with open(out, "w") as f:
        f.write("contig_id\tlength\talignment_rate\n")
        for contig, values in contigs.items():
            row = [contig, str(values["len"]), f"{values['hits']}/{len(genomes)}"]
            f.write("\t".join(row) + "\n")
    print(f"  contig features: {out}")
    print(
        "\u001b[1m"
        + "\n• Identifying contigs with no conspecific alignments"
        + "\u001b[0m"
    )
    flagged = flag_contigs(args, contigs)
    out = f"{args['tmp_dir']}/flagged_contigs"
    with open(out, "w") as f:
        for contig in flagged:
            f.write(contig + "\n")
    print(f"  {len(flagged)} flagged contigs: {out}")



# -*- coding: utf-8 -*-
#
#   This file is part of the magpurify package, available at:
#   https://github.com/snayfach/MAGpurify
#
#   Magpurify is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import itertools
import os
import sys
import Bio.Seq
import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from magpurify import utilities


def fetch_args(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="tetra-freq")
    parser.add_argument("fna", type=str, help="Path to input genome in FASTA format")
    parser.add_argument(
        "out", type=str, help="Output directory to store results and intermediate files",
    )
    parser.add_argument("--cutoff", type=float, default=0.06, help="Cutoff")
    parser.add_argument(
        "--weighted-mean",
        action="store_true",
        help="Compute the mean weighted by the contig length"
    )


def init_kmers():
    tetra = {}
    for i in itertools.product("ACGT", repeat=4):
        kmer_fwd = "".join(i)
        kmer_rev = utilities.reverse_complement(kmer_fwd)
        if kmer_fwd in tetra or kmer_rev in tetra:
            continue
        else:
            tetra[kmer_fwd] = 0
    return tetra


class Contig:
    def __init__(self):
        pass


def main(args):
    utilities.add_tmp_dir(args)
    utilities.check_input(args)
    utilities.check_dependencies(["blastn"])

    print("\u001b[1m" + "• Counting tetranucleotides" + "\u001b[0m")
    # init data
    contigs = {}
    contig_length_list = []
    for id, seq in utilities.parse_fasta(args["fna"]):
        contig = Contig()
        contig.id = id
        contig.seq = str(seq)
        contig.kmers = init_kmers()
        contigs[id] = contig
        contig_length_list.append(len(seq))

    # count kmers
    for contig in contigs.values():
        for i in range(len(contig.seq) - 3):
            kmer_fwd = contig.seq[i : i + 4]
            if kmer_fwd in contig.kmers:
                contig.kmers[kmer_fwd] += 1
            else:
                kmer_rev = utilities.reverse_complement(kmer_fwd)
                contig.kmers[kmer_rev] += 1

    print("\u001b[1m" + "\n• Normalizing counts" + "\u001b[0m")
    for contig in contigs.values():
        total = float(sum(contig.kmers.values()))
        for kmer, count in contig.kmers.items():
            contig.kmers[kmer] = 100 * count / total if total > 0 else 0.0
    print("\u001b[1m" + "\n• Performing PCA" + "\u001b[0m")
    df = pd.DataFrame(dict([(c.id, c.kmers) for c in contigs.values()]))
    pca = PCA(n_components=1)
    pca.fit(df)
    pc1 = pca.components_[0]
    if args["weighted_mean"]:
        print(
            "\u001b[1m"
            + "\n• Computing per-contig deviation from the weighted mean along the first principal component"
            + "\u001b[0m"
        )
        reference_pc = np.average(pc1, weights=contig_length_list)
    else:
        print(
            "\u001b[1m"
            + "\n• Computing per-contig deviation from the mean along the first principal component"
            + "\u001b[0m"
        )
        reference_pc = np.average(pc1)
    for contig_id, contig_pc in zip(list(df.columns), pc1):
        contigs[contig_id].pc = contig_pc
        contigs[contig_id].values = {}
        contigs[contig_id].values["delta"] = abs(contig_pc - reference_pc)

    print("\u001b[1m" + "\n• Identifying outlier contigs" + "\u001b[0m")
    flagged = [
        contig.id
        for contig in contigs.values()
        if contig.values["delta"] > args["cutoff"]
    ]

    out = f"{args['tmp_dir']}/flagged_contigs"
    print(f"  {len(flagged)} flagged contigs: {out}")
    with open(out, "w") as f:
        for contig in flagged:
            f.write(contig + "\n")



# -*- coding: utf-8 -*-
#
#   This file is part of the magpurify package, available at:
#   https://github.com/snayfach/MAGpurify
#
#   Magpurify is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import os
import sys
import numpy as np
from Bio import SeqUtils
from magpurify import utilities


def fetch_args(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="gc-content")
    parser.add_argument(
        "fna",
        type=str,
        help="Path to input genome in FASTA format"
    )
    parser.add_argument(
        "out",
        type=str,
        help="Output directory to store results and intermediate files",
    )
    parser.add_argument(
        "--cutoff", type=float, default=15.75, help="Cutoff"
    )
    parser.add_argument(
        "--weighted-mean",
        action="store_true",
        help="Compute the mean weighted by the contig length"
    )


class Contig:
    def __init__(self):
        pass


def main(args):
    utilities.add_tmp_dir(args)
    utilities.check_input(args)
    if args["weighted_mean"]:
        print("\u001b[1m" + "• Computing weighted mean contig GC content" + "\u001b[0m")
    else:
        print("\u001b[1m" + "• Computing mean contig GC content" + "\u001b[0m")
    contigs = {}
    contig_length_list = []
    for id, seq in utilities.parse_fasta(args["fna"]):
        contig = Contig()
        contig.id = id
        contig.seq = str(seq)
        contig.gc = round(SeqUtils.GC(seq), 2)
        contigs[id] = contig
        contig_length_list.append(len(seq))
    if args["weighted_mean"]:
        print("\u001b[1m" + "\n• Computing per-contig deviation from weighted mean" + "\u001b[0m")
        reference = np.average([c.gc for c in contigs.values()], weights=contig_length_list)
    else:
        print("\u001b[1m" + "\n• Computing per-contig deviation from mean" + "\u001b[0m")
        reference = np.average([c.gc for c in contigs.values()])
    for contig in contigs.values():
        contig.values = {}
        contig.values["delta"] = abs(contig.gc - reference)
    print("\u001b[1m" + "\n• Identifying outlier contigs" + "\u001b[0m")
    flagged = [
        contig.id
        for contig in contigs.values()
        if contig.values["delta"] > args["cutoff"]
    ]

    out = f"{args['tmp_dir']}/flagged_contigs"
    print(f"  {len(flagged)} flagged contigs: {out}")
    with open(out, "w") as f:
        for contig in flagged:
            f.write(contig + "\n")



# -*- coding: utf-8 -*-
#
#   This file is part of the magpurify package, available at:
#   https://github.com/snayfach/MAGpurify
#
#   Magpurify is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import sys
import numpy as np
import pandas as pd
from magpurify import utilities


def fetch_args(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="coverage")
    parser.add_argument(
        "fna",
        type=str,
        help="Path to input genome in FASTA format")
    parser.add_argument(
        "out",
        type=str,
        help="Output directory to store results and intermediate files",
    )
    parser.add_argument(
        "bams",
        nargs="+",
        type=str,
        help="Path to input sorted BAM file(s)",
    )
    parser.add_argument(
        "--max-deviation",
        type=float,
        default=5.0,
        help="Contigs with coverage greater than [max-deviation * mean coverage] or less than [(1/max-deviation) * mean coverage] will be flagged as outliers",
    )
    parser.add_argument(
        "--weighted-mean",
        action="store_true",
        help="Compute the mean weighted by the contig length"
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=1,
        help="Number of CPUs to use",
    )


def main(args):
    utilities.add_tmp_dir(args)
    utilities.check_input(args)
    utilities.check_dependencies(["coverm"])
    print("\u001b[1m" + "• Computing contig coverage" + "\u001b[0m")
    utilities.run_coverm(args["bams"], args["tmp_dir"], args["threads"])
    coverage_df = pd.read_csv(f"{args['tmp_dir']}/coverage.tsv", sep="\t", index_col=0)
    contig_id_list = []
    contig_length_list = []
    for id, seq in utilities.parse_fasta(args["fna"]):
        contig_id_list.append(id)
        contig_length_list.append(len(seq))
    contig_coverage_df = coverage_df.loc[contig_id_list]
    largest_mean_coverage_sample = contig_coverage_df.mean(axis=0).idxmax()
    if contig_coverage_df.shape[1] > 1:
        print(
            "\u001b[1m"
            + f"\n• Sample being used for outlier detection: {largest_mean_coverage_sample.split()[0]}"
            + "\u001b[0m"
        )
    contig_coverage_df = contig_coverage_df.loc[:, largest_mean_coverage_sample]
    if contig_coverage_df.mean() < 1:
        sys.exit(
            "\nError: The average coverage is less than 1 in all the supplied BAM files"
        )
    if args["weighted_mean"]:
        print("\u001b[1m" + "\n• Computing per-contig deviation from the weighted mean coverage" + "\u001b[0m")
        reference = np.average(contig_coverage_df.values, weights=contig_length_list)
    else:
        print("\u001b[1m" + "\n• Computing per-contig deviation from the mean coverage" + "\u001b[0m")
        reference = contig_coverage_df.mean()
    outliers = ((contig_coverage_df / reference) >= args["max_deviation"]) | (
        (contig_coverage_df / reference) <= 1 / args["max_deviation"]
    )
    print("\u001b[1m" + "\n• Identifying outlier contigs" + "\u001b[0m")
    flagged = contig_coverage_df.loc[outliers].index.tolist()
    out = f"{args['tmp_dir']}/flagged_contigs"
    print(f"  {len(flagged)} flagged contigs: {out}")
    with open(out, "w") as f:
        for contig in flagged:
            f.write(contig + "\n")




# -*- coding: utf-8 -*-
#
#   This file is part of the magpurify package, available at:
#   https://github.com/snayfach/MAGpurify
#
#   Magpurify is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import collections
import copy
import operator
import os
import sys
from magpurify import utilities

ranks = ["k", "p", "c", "o", "f", "g", "s"]
rank_names = {
    "k": "kingdom",
    "p": "phylum",
    "c": "class",
    "o": "order",
    "f": "family",
    "g": "genus",
    "s": "species",
}


def fetch_args(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="clade-markers")
    parser.add_argument("fna", type=str, help="Path to input genome in FASTA format")
    parser.add_argument(
        "out", type=str, help="Output directory to store results and intermediate files",
    )
    parser.add_argument(
        "--db",
        type=str,
        help="Path to reference database. By default, the MAGPURIFY environmental variable is used",
    )
    parser.add_argument(
        "--exclude_clades",
        nargs="+",
        type=str,
        help="List of clades to exclude (ex: s__Variovorax_sp_CF313)",
    )
    parser.add_argument(
        "--min_bin_fract",
        type=float,
        default=0.6,
        help="Min fraction of bin length supported by contigs that agree with consensus taxonomy",
    )
    parser.add_argument(
        "--min_contig_fract",
        type=float,
        default=0.75,
        help="Min fraction of classified contig length that agree with consensus taxonomy",
    )
    parser.add_argument(
        "--min_gene_fract",
        type=float,
        default=0.0,
        help="Min fraction of classified genes that agree with consensus taxonomy",
    )
    parser.add_argument(
        "--min_genes",
        type=float,
        default=None,
        help="Min number of genes that agree with consensus taxonomy (default=rank-specific-cutoffs)",
    )
    parser.add_argument(
        "--lowest_rank",
        choices=["s", "g", "f", "o", "c", "p", "k"],
        help="Lowest rank for bin classification",
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=1,
        help="Number of CPUs to use",
    )


def read_ref_taxonomy(db_dir):
    ref_taxonomy = {}
    inpath = f"{db_dir}/clade-markers/taxonomy.tsv"
    for line in open(inpath):
        ref_id, taxonomy = line.rstrip().split()
        ref_taxonomy[ref_id] = taxonomy
    return ref_taxonomy


def flag_contigs(contigs, bin):
    for bin_taxon in bin.taxonomy:
        bin_rank = bin_taxon.split("__")[0]
        for contig in contigs.values():
            contig_taxa = [
                gene.taxa[bin_rank] for gene in contig.genes if gene.taxa[bin_rank]
            ]
            # contig has no gene hits at rank
            if len(contig_taxa) == 0:
                continue
            # flag contig as discordant
            elif bin_taxon not in contig_taxa:
                contig.flagged = True
            # contig already flagged as discordant at higher rank
            elif contig.flagged:
                continue
            # flag contig as concordant
            # this is debatable...contig could be flagged as concordant,
            # but at a higher rank than the bin in annotated to
            else:
                contig.flagged = False


class Gene:
    def __init__(self):
        self.id = None
        self.aln = None
        self.taxa = dict([(rank, None) for rank in ranks])


class Contig:
    def __init__(self):
        self.id = None
        self.length = None
        self.genes = []
        self.flagged = None
        self.cons_taxa = dict([(rank, None) for rank in ranks])

    def classify(self):
        for rank in ranks:
            taxa = [g.taxa[rank] for g in self.genes if g.taxa[rank]]
            if len(taxa) > 0:
                counts = collections.Counter(taxa).items()
                self.cons_taxa[rank] = sorted(
                    counts, key=operator.itemgetter(1), reverse=True
                )[0][0]


class Bin:
    def __init__(self):
        self.cons_taxon = None
        self.bin_fract = None
        self.contig_fract = None
        self.gene_fract = None
        self.taxonomy = [None]
        self.tagged_length = 0
        self.tagged_genes = 0

    def classify(
        self,
        contigs,
        min_bin_fract,
        min_contig_fract,
        min_gene_fract,
        min_genes,
        lowest_rank,
    ):
        for rank in ranks:
            count_genes = collections.defaultdict(int)
            for contig in contigs.values():
                for gene in contig.genes:
                    if gene.taxa[rank]:
                        count_genes[gene.taxa[rank]] += 1
            count_length = collections.defaultdict(int)
            for contig in contigs.values():
                contig_taxon = contig.cons_taxa[rank]
                if contig_taxon is not None:
                    count_length[contig_taxon] += contig.length
            if sum(count_length.values()) > 0:
                cons_taxon = sorted(
                    count_length.items(), key=operator.itemgetter(1), reverse=True
                )[0][0]
                gene_fract = count_genes[cons_taxon] / float(sum(count_genes.values()))
                contig_fract = count_length[cons_taxon] / float(
                    sum(count_length.values())
                )
                bin_fract = sum(count_length.values()) / sum(
                    c.length for c in contigs.values()
                )
            else:
                cons_taxon = "NA"
                gene_fract = 0.0
                contig_fract = 0.0
                bin_fract = 0.0
            if bin_fract < min_bin_fract:
                continue
            elif contig_fract < min_contig_fract:
                continue
            elif gene_fract < min_gene_fract:
                continue
            elif count_genes[cons_taxon] < max(min_genes[rank], 1):
                continue
            else:
                self.cons_taxon = cons_taxon
                self.gene_fract = gene_fract
                self.contig_fract = contig_fract
                self.bin_fract = bin_fract
                self.tagged_genes = sum(count_genes.values())
                self.tagged_length = sum(count_length.values())
            if lowest_rank and rank == lowest_rank:
                break


def main(args):
    utilities.add_tmp_dir(args)
    utilities.check_input(args)
    utilities.check_database(args)
    print("\u001b[1m" + "• Reading database info" + "\u001b[0m")
    ref_taxonomy = read_ref_taxonomy(args["db"])
    taxon_to_taxonomy = {}
    for taxonomy in set(ref_taxonomy.values()):
        for taxon in taxonomy.split("|"):
            taxon_to_taxonomy[taxon] = taxonomy
    min_pid = {"k": 57, "p": 77, "c": 82, "o": 86, "f": 87, "g": 91, "s": 96}
    if args["min_genes"] is not None:
        args["min_genes"] = dict([(r, args["min_genes"]) for r in ranks])
    else:
        args["min_genes"] = {
            "k": 237,
            "p": 44,
            "c": 30,
            "o": 24,
            "f": 22,
            "g": 20,
            "s": 19,
        }
    print("\u001b[1m" + "\n• Calling genes with Prodigal" + "\u001b[0m")
    utilities.run_prodigal(args["fna"], args["tmp_dir"])
    print(f"  all genes: {args['tmp_dir']}/genes.[ffn|faa]")
    print(
        "\u001b[1m"
        + "\n• Performing pairwise alignment of genes against MetaPhlan2 database of clade-specific genes"
        + "\u001b[0m"
    )
    utilities.run_lastal(args["db"], args["tmp_dir"], args["threads"])
    print(f"  alignments: {args['tmp_dir']}/genes.m8")

    print("\u001b[1m" + "\n• Finding top hits to database" + "\u001b[0m")
    genes = {}
    for aln in utilities.parse_last(args["tmp_dir"] + "/genes.m8"):
        # clade exclusion
        ref_taxa = ref_taxonomy[aln["tid"]].split("|")
        if args["exclude_clades"] and any(
            taxon in ref_taxa for taxon in args["exclude_clades"]
        ):
            continue
        # initialize gene
        if aln["qid"] not in genes:
            genes[aln["qid"]] = Gene()
            genes[aln["qid"]].id = aln["qid"]
            genes[aln["qid"]].contig_id = aln["qid"].rsplit("_", 1)[0]

        # get top alignments
        if genes[aln["qid"]].aln is None:
            genes[aln["qid"]].aln = aln
            genes[aln["qid"]].ref_taxa = ref_taxa
        elif float(aln["score"]) > float(genes[aln["qid"]].aln["score"]):
            genes[aln["qid"]].ref_taxa = ref_taxa
    print("  %s genes with a database hit" % len(genes))
    print("\u001b[1m" + "\n• Classifying genes at each taxonomic rank" + "\u001b[0m")
    counts = {}
    for gene in genes.values():
        for ref_taxon in gene.ref_taxa:
            rank = ref_taxon.split("__")[0]
            if rank not in counts:
                counts[rank] = 0
            if rank == "t":
                continue
            elif float(gene.aln["pid"]) < min_pid[rank]:
                continue
            elif gene.aln["qcov"] < 0.4:
                continue
            elif gene.aln["tcov"] < 0.4:
                continue
            gene.taxa[rank] = ref_taxon
            counts[rank] += 1
    for rank in ranks:
        print(f"  {rank_names[rank]}: {counts[rank]} classified genes")
    print("\u001b[1m" + "\n• Taxonomically classifying contigs" + "\u001b[0m")
    contigs = {}
    for id, seq in utilities.parse_fasta(args["fna"]):
        contigs[id] = Contig()
        contigs[id].id = id
        contigs[id].length = len(seq)
    # aggregate hits by contig
    for gene in genes.values():
        contigs[gene.contig_id].genes.append(gene)
    # classify contigs at each level
    for contig in contigs.values():
        contig.classify()
    # summarize
    counts = {}
    for contig in contigs.values():
        for rank, taxon in contig.cons_taxa.items():
            if rank not in counts:
                counts[rank] = 0
            if taxon is not None:
                counts[rank] += 1
    print("  total contigs: %s" % len(contigs))
    for rank in ranks:
        print(f"  {rank_names[rank]}: {counts[rank]} classified contigs")

    print("\u001b[1m" + "\n• Taxonomically classifying genome" + "\u001b[0m")
    bin = Bin()
    bin.classify(
        contigs,
        args["min_bin_fract"],
        args["min_contig_fract"],
        args["min_gene_fract"],
        args["min_genes"],
        args["lowest_rank"],
    )
    print(f"  consensus taxon: {bin.cons_taxon}")
    print("\u001b[1m" + "\n• Identifying taxonomically discordant contigs" + "\u001b[0m")
    if bin.cons_taxon is not None:
        bin.rank_index = (
            taxon_to_taxonomy[bin.cons_taxon].split("|").index(bin.cons_taxon)
        )
        bin.taxonomy = taxon_to_taxonomy[bin.cons_taxon].split("|")[
            0 : bin.rank_index + 1
        ]
        flag_contigs(contigs, bin)
    flagged = [contig.id for contig in contigs.values() if contig.flagged]
    out = f"{args['tmp_dir']}/flagged_contigs"
    print(f"  {len(flagged)} flagged contigs: {out}")
    with open(out, "w") as f:
        for contig in flagged:
            f.write(contig + "\n")




# -*- coding: utf-8 -*-
#
#   This file is part of the magpurify package, available at:
#   https://github.com/snayfach/MAGpurify
#
#   Magpurify is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import os
import sys
import textwrap
from operator import itemgetter
from magpurify import utilities


def fetch_args(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="clean-bin")
    parser.add_argument(
        "fna",
        type=str,
        help="Path to input genome in FASTA format"
    )
    parser.add_argument(
        "out",
        type=str,
        help="Output directory to store results and intermediate files",
    )
    parser.add_argument(
        "out_fna",
        type=str,
        help="Path to the output FASTA file",
    )


def main(args):
    utilities.check_input(args)
    print("\u001b[1m" + "• Reading genome bin" + "\u001b[0m")
    bin = {id: seq for id, seq in utilities.parse_fasta(args["fna"])}
    bin_length = round(sum(len(_) for _ in bin.values()) / 1000, 2)
    print(f"  genome length: {len(bin)} contigs, {bin_length} Kbp")
    print("\u001b[1m" + "\n• Reading flagged contigs" + "\u001b[0m")
    flagged_contigs = []
    programs = [
        "phylo-markers",
        "clade-markers",
        "conspecific",
        "tetra-freq",
        "gc-content",
        "coverage",
        "known-contam",
    ]
    for program in programs:
        path = f"{args['out']}/{program}/flagged_contigs"
        if not os.path.exists(path):
            print(f"  {program}: no output file found")
        else:
            contigs = [_.rstrip() for _ in open(path)]
            bases = round(sum(len(bin[id]) for id in contigs) / 1000, 2)
            flagged_contigs += contigs
            print(f"  {program}: {len(contigs)} contigs, {bases} Kbp")
    flagged_contigs = list(set(flagged_contigs))
    flagged_length = round(sum(len(bin[id]) for id in flagged_contigs) / 1000, 2)
    print("\u001b[1m" + "\n• Removing flagged contigs" + "\u001b[0m")
    clean = bin.copy()
    for id in flagged_contigs:
        del clean[id]
    clean_length = round(sum(len(_) for _ in clean.values()) / 1000, 2)
    print(f"  removed: {len(flagged_contigs)} contigs, {flagged_length} Kbp")
    print(f"  remains: {len(clean)} contigs, {clean_length} Kbp")
    with open(args['out_fna'], "w") as f:
        for id, seq in clean.items():
            f.write(">" + id + "\n" + textwrap.fill(seq, 70) + "\n")
    print(f"  cleaned bin: {args['out_fna']}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   This file is part of the magpurify package, available at:
#   https://github.com/snayfach/MAGpurify
#
#   Magpurify is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.

from setuptools import find_packages, setup

setup(
    name="magpurify",
    version="2.1.2",
    packages=find_packages(),
    license="GNU General Public License v3.0",
    description="Identify and remove incorrectly binned contigs from metagenome-assembled genomes.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "biopython",
        "numpy",
        "pandas",
        "scikit-learn",
    ],
    python_requires=">=3.6",
    entry_points={"console_scripts": ["magpurify=magpurify.cli:cli"]},
    url="https://github.com/snayfach/MAGpurify",
    keywords=[
        "bioinformatics",
        "metagenomics",
        "metagenome-assembled genomes"
    ],
    author="Stephen Nayfach, Antonio Pedro Camargo",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python :: 3",
    ],
)



# -*- coding: utf-8 -*-
#
#   This file is part of the magpurify package, available at:
#   https://github.com/snayfach/MAGpurify
#
#   Magpurify is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import os
import sys
from magpurify import utilities


def fetch_args(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="known-contam")
    parser.add_argument(
        "fna",
        type=str,
        help="Path to input genome in FASTA format"
    )
    parser.add_argument(
        "out",
        type=str,
        help="Output directory to store results and intermediate files",
    )
    parser.add_argument(
        "--db",
        type=str,
        help="Path to reference database. By default, the IMAGEN_DB environmental variable is used",
    )
    parser.add_argument(
        "--pid",
        type=float,
        default=98,
        help="Minimum %% identity to reference",
    )
    parser.add_argument(
        "--evalue",
        type=float,
        default=1e-5,
        help="Maximum evalue"
    )
    parser.add_argument(
        "--qcov",
        type=float,
        default=25,
        help="Minimum percent query coverage",
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=1,
        help="Number of CPUs to use",
    )


def run_blastn(query, db, out, threads, qcov=25, pid=98, evalue=1e-5):
    cmd = "blastn "
    cmd += f"-query {query} "
    cmd += f"-db {db} "
    cmd += f"-out {out} "
    cmd += "-outfmt '6 std qlen slen' "
    cmd += "-max_target_seqs 1 "
    cmd += "-max_hsps 1 "
    cmd += f"-qcov_hsp_perc {qcov} "
    cmd += f"-perc_identity {pid} "
    cmd += f"-evalue {evalue} "
    cmd += f"-num_threads {threads} "
    utilities.run_process(cmd)


def main(args):
    utilities.add_tmp_dir(args)
    utilities.check_input(args)
    utilities.check_dependencies(["blastn"])
    utilities.check_database(args)
    utilities.add_tmp_dir(args)
    print("\u001b[1m" + "• Searching database with BLASTN" + "\u001b[0m")
    for target in ["hg38", "phix"]:
        db = f"{args['db']}/known-contam/{target}/{target}"
        out = f"{args['tmp_dir']}/{target}.m8"
        run_blastn(
            args["fna"],
            db,
            out,
            args["threads"],
            args["qcov"],
            args["pid"],
            args["evalue"],
        )
    print("\u001b[1m" + "\n• Identifying contigs with hits to database" + "\u001b[0m")
    flagged = set([])
    for target in ["hg38", "phix"]:
        out = f"{args['tmp_dir']}/{target}.m8"
        for r in utilities.parse_blast(out):
            flagged.add(r["qname"])
    flagged = list(flagged)
    out = f"{args['tmp_dir']}/flagged_contigs"
    print(f"  {len(flagged)} flagged contigs: {out}")
    with open(out, "w") as f:
        for contig in flagged:
            f.write(contig + "\n")
